"""
MIT License

Copyright (c) 2021 Qualichat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
import sys
import platform
from typing import List, Tuple
from argparse import ArgumentParser, Namespace, Action

import plotly # type: ignore
import spacy
from rich import print
from rich.logging import RichHandler
from spacy.cli.download import download

import qualichat
from ._partials import *
from qualichat.frames import BaseFrame
from qualichat.utils import config, log


ascii = '''
                                ████   ███           █████                 █████   
                               ░░███  ░░░           ░░███                 ░░███    
  ████████ █████ ████  ██████   ░███  ████   ██████  ░███████    ██████   ███████  
 ███░░███ ░░███ ░███  ░░░░░███  ░███ ░░███  ███░░███ ░███░░███  ░░░░░███ ░░░███░   
░███ ░███  ░███ ░███   ███████  ░███  ░███ ░███ ░░░  ░███ ░███   ███████   ░███    
░███ ░███  ░███ ░███  ███░░███  ░███  ░███ ░███  ███ ░███ ░███  ███░░███   ░███ ███
░░███████  ░░████████░░████████ █████ █████░░██████  ████ █████░░████████  ░░█████ 
 ░░░░░███   ░░░░░░░░  ░░░░░░░░ ░░░░░ ░░░░░  ░░░░░░  ░░░░ ░░░░░  ░░░░░░░░    ░░░░░  
     ░███                                                                          
     █████                                                                         
    ░░░░░                                                                          
'''


def show_version() -> None:
    entries: List[str] = []
    formatter = 'v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'

    entries.append(f'- Python {formatter.format(sys.version_info)}')
    entries.append(f'- Qualichat {formatter.format(qualichat.version_info)}')
    entries.append(f'- Plotly {plotly.__version__}')
    entries.append(f'- spaCy {spacy.__version__}') # type: ignore

    uname = platform.uname()
    entries.append('- System info: {0.system} {0.release}'.format(uname))

    print('\n'.join(entries))


def core(parser: ArgumentParser, args: Namespace) -> None:
    if args.version:
        show_version()


def loadchat(parser: ArgumentParser, args: Namespace) -> None:
    debug = args.debug
    qc = qualichat.load_chats(*args.paths, debug=debug)

    print(ascii)
    log('info', 'Welcome to Qualichat.')

    while True:
        choices = list(qc.frames.keys())
        frame_name: str = select('Choose a frame:', choices).ask()

        if not frame_name:
            return log('error', 'Operation canceled. Aborting.')

        frame: BaseFrame = qc.frames[frame_name]

        choices = list(frame.charts.keys())
        names = checkbox('Choose your charts:', choices).ask()

        if not names:
            return log('error', 'No charts were selected. Aborting.')

        for name in names:
            frame.charts[name](qc.chats)

        log('info', 'Restarting menu...')


def setup(parser: ArgumentParser, args: Namespace) -> None:
    api_key = password('Enter your Google API key:').ask()

    config['google_api_key'] = api_key
    config.save()

    with progress_bar(transient=True) as progress:
        progress.add_task('[green]Downloading spaCy models', start=False)
        download('pt_core_news_md', False, False, '-q')

    print('\n[green]✔ You can now use Qualichat.[/green]')


def add_loadchat_args(subparser: Action) -> None:
    parser_help = 'start an interactive session with Qualichat'
    parser = subparser.add_parser('load', help=parser_help) # type: ignore
    parser.set_defaults(func=loadchat) # type: ignore

    path_arg_help = 'the paths for the chats'
    parser.add_argument('paths', help=path_arg_help, nargs='+') # type: ignore

    debug_args = ('-d', '--debug')
    debug_args_help = 'set the logging level to debug'
    parser.add_argument( # type: ignore
        *debug_args, help=debug_args_help, action='store_true'
    )


def add_setup_args(subparser: Action) -> None:
    parser_help = 'setup internal libraries'
    parser = subparser.add_parser('setup', help=parser_help) # type: ignore
    parser.set_defaults(func=setup) # type: ignore


def parse_args() -> Tuple[ArgumentParser, Namespace]:
    desc = 'Tools for using Qualichat.'
    parser = ArgumentParser(prog='qualichat', description=desc)

    help = 'show the library version'
    parser.add_argument('-v', '--version', action='store_true', help=help)
    parser.set_defaults(func=core)

    subparser = parser.add_subparsers(dest='subcommand', title='subcommands')
    add_loadchat_args(subparser)
    add_setup_args(subparser)

    return (parser, parser.parse_args())


def main() -> None:
    handler = RichHandler(
        omit_repeated_times=False, show_path=False, markup=True
    )

    logger = logging.getLogger('qualichat')
    logger.addHandler(handler)

    parser, args = parse_args()
    args.func(parser, args)


if __name__ == '__main__':
    main()



"""
MIT License
Copyright (c) 2021 Qualichat
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import base64
from io import BytesIO
from collections import defaultdict
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from wordcloud import WordCloud # type: ignore
from plotly.graph_objs import Scatter, Figure # type: ignore
from plotly.subplots import make_subplots # type: ignore
from pandas import DataFrame

from .chat import Chat
from .models import Message
from .utils import log
from ._partials import progress_bar, select, checkbox


__all__ = ('generate_wordcloud', 'keys')


def generate_chart(
    dataframes: Dict[Chat, DataFrame],
    *,
    bars: Optional[List[str]] = None,
    lines: Optional[List[str]] = None,
    title: Optional[str] = None,
) -> None:
    """
    """
    if bars is None:
        bars = []

    if lines is None:
        lines = []

    specs = [[{'secondary_y': True}]]
    fig = make_subplots(specs=specs) # type: ignore

    buttons: List[Dict[str, Any]] = []
    visible = True

    for i, (chat, dataframe) in enumerate(dataframes.items()):
        dataframe = dataframe.sort_values( # type: ignore
            by=bars + lines, ascending=False
        )
        index = list(dataframe.index) # type: ignore

        button: Dict[str, Any] = {}
        button['label'] = chat.filename
        button['method'] = 'update'

        args: List[Union[Dict[str, Any], List[Dict[str, Any]]]] = []

        visibility: List[bool] = []
        for j in range(len(dataframes)):
            for _ in range(len(bars + lines)):
                visibility.append(i == j)

        args.append({'visible': visibility})
        args.append({'title': {'text': f'{title} ({chat.filename})'}})

        button['args'] = args
        buttons.append(button)

        for bar in bars:
            filtered = getattr(dataframe, bar)
            options = dict(x=index, y=list(filtered), name=bar, visible=visible) # type: ignore
            fig.add_bar(**options) # type: ignore

        for line in lines:
            filtered = getattr(dataframe, line)
            scatter = Scatter(x=index, y=list(filtered), name=line, visible=visible) # type: ignore
            fig.add_trace(scatter, secondary_y=True) # type: ignore

        if visible is True:
            visible = False

    updatemenus = [{'buttons': buttons, 'active': 0}]
    fig.update_layout(updatemenus=updatemenus) # type: ignore

    fig.update_xaxes(rangeslider_visible=True) # type: ignore
    fig.show() # type: ignore


def generate_table(
    tables: Dict[Chat, DataFrame],
    *,
    columns: Optional[List[str]] = None,
    title: str
):
    """
    """
    if columns is None:
        columns = []

    fig = make_subplots() # type: ignore
    buttons: List[Dict[str, Any]] = []

    for chat, dataframe in tables.items():
        button: Dict[str, Any] = {}
        button['label'] = chat.filename
        button['method'] = 'update'

        args: List[Dict[str, Any]] = []
        values: List[Any] = []

        for column in columns:
            values.append(dataframe[column].to_list()) # type: ignore

        args.append({'cells': {'values': values}, 'header': {'values': columns}})
        args.append({'title': {'text': f'{title} ({chat.filename})'}})

        button['args'] = args
        buttons.append(button)

    header: Dict[str, List[str]] = {'values': []}
    cells: Dict[str, List[Any]] = {'values': []}

    fig.add_table(header=header, cells=cells) # type: ignore

    updatemenus = [{'buttons': buttons, 'active': 0}]
    fig.update_layout(updatemenus=updatemenus) # type: ignore

    fig.show() # type: ignore


def _sort_by_time(chats: List[Chat]) -> Dict[Chat, Dict[str, List[Message]]]:
    ret: Dict[Chat, Dict[str, List[Message]]] = {}

    def sort(messages: List[Message]):
        data: DefaultDict[str, List[Message]] = defaultdict(list)

        with progress_bar() as progress:
            for m in progress.track(messages, description='Sorting...'):
                data[m.created_at.strftime('%B %Y')].append(m)

        choices = ['All', 'Choose an epoch']
        message = f'[{chat.filename}] Which messages should be selected?'
        selected = select(message, choices).ask()

        if selected == 'All':
            return dict(data)

        choices = list(data.keys())
        if not (epochs := checkbox('Choose an epoch:', choices).ask()):
            raise KeyError()

        return {epoch: data[epoch] for epoch in epochs}

    for chat in chats:
        ret[chat] = sort(chat.messages)

    return ret


def _sort_by_actor(chats: List[Chat]) -> Dict[Chat, Dict[str, List[Message]]]:
    ret: Dict[Chat, Dict[str, List[Message]]] = {}

    def sort(messages: List[Message]) -> Dict[str, List[Message]]:
        data: DefaultDict[str, List[Message]] = defaultdict(list)

        with progress_bar() as progress:
            for m in progress.track(messages, description='Tracking...'):
                data[m.actor.display_name].append(m)

        choices = ['All', 'Choose a specific actor']
        message = f'[{chat.filename}] Which actors should be selected?'
        selected = select(message, choices).ask()

        if selected == 'All':
            return dict(data)

        choices = list(data.keys())
        if not (actors := checkbox('Choose an actor:', choices).ask()):
            raise KeyError()

        if len(actors) != 1:
            ret: Dict[str, List[Message]] = {}
            others: List[Message] = []

            for actor in data:
                if actor not in actors:
                    others.extend(data[actor])
                else:
                    ret[actor] = data[actor]

            ret['Others'] = others
            return ret

        actor = actors[0]
        messages = data[actor]

        new_data: DefaultDict[str, List[Message]] = defaultdict(list)

        with progress_bar() as progress:
            for m in progress.track(messages, description='Sorting...'):
                new_data[m.created_at.strftime('%B %Y')].append(m)

        return dict(new_data)

    for chat in chats:
        ret[chat] = sort(chat.messages)

    return ret


def generate_treemap(dataframes: Dict[Chat, DataFrame], **kwargs: Any):
    """
    """
    title = kwargs.pop('title', None)
    have_parents = kwargs.pop('have_parents', True)

    specs = [[{'secondary_y': True}]]
    fig = make_subplots(specs=specs) # type: ignore

    buttons: List[Dict[str, Any]] = []
    visible = True

    for i, (chat, dataframe) in enumerate(dataframes.items()):
        index = list(dataframe.index) # type: ignore

        button: Dict[str, Any] = {}
        button['label'] = chat.filename
        button['method'] = 'update'

        args: List[Union[Dict[str, Any], List[Dict[str, Any]]]] = []

        visibility: List[bool] = []
        for j in range(len(dataframes)):
            for _ in range(len(list(dataframe))): # type: ignore
                visibility.append(i == j)

        args.append({'visible': visibility})
        args.append({'title': {'text': f'{title} ({chat.filename})'}})

        button['args'] = args
        buttons.append(button)

        labels: List[str] = []
        values: List[int] = []
        parents: List[str] = []

        if have_parents:
            for index, row in dataframe.iterrows():
                labels.append(index) # type: ignore
                values.append(0)
                parents.append('')

                for column, value in zip(list(dataframe), row): # type: ignore
                    labels.append(f'{column} ({index})')
                    values.append(value) # type: ignore
                    parents.append(index) # type: ignore
        else:
            for i, row in dataframe.iterrows():
                value = row.values[0] # type: ignore

                labels.append(i) # type: ignore
                values.append(value) # type: ignore
                parents.append('')

        fig.add_treemap( # type: ignore
            labels=labels, values=values, parents=parents, visible=visible
        )

        if visible is True:
            visible = False

    fig.show() # type: ignore


def generate_wordcloud(wordclouds: Dict[Chat, WordCloud], *, title: str):
    """
    """
    fig = Figure() # type: ignore
    buttons: List[Dict[str, Any]] = []

    for i, (chat, wordcloud) in enumerate(wordclouds.items()):
        button: Dict[str, Any] = {}
        button['label'] = chat.filename
        button['method'] = 'update'

        args: List[Dict[str, Any]] = []

        visibility: List[bool] = []
        for j in range(len(wordclouds)):
            visibility.append(i == j)

        args.append({'visible': visibility})
        args.append({'title': {'text': f'{title} ({chat.filename})'}})

        button['args'] = args

        buffer = BytesIO()
        image = wordcloud.to_image() # type: ignore

        image.save(buffer, format='png')
        buffer.seek(0)
        encoded_file = base64.b64encode(buffer.read()).decode('utf-8')

        source = f'data:image/png;base64, {encoded_file}'
        fig.add_image(source=source) # type: ignore

        buttons.append(button)

    updatemenus = [{'buttons': buttons, 'active': 0}]
    fig.update_layout(updatemenus=updatemenus) # type: ignore

    fig.show() # type: ignore


def keys(func: Callable[..., None]):
    """
    """

    # Hack to avoid circular imports.
    from .frames import BaseFrame

    def decorator(self: BaseFrame, chats: List[Chat]) -> None:
        modes = {'By Time': _sort_by_time, 'By Actor': _sort_by_actor}
        choices = list(modes.keys())

        name = select('Choose your mode:', choices).ask()
        sorter_type = modes[name]

        try:
            sorted_messages = sorter_type(chats)
        except (KeyError, TypeError):
            return log('error', 'Option not selected. Aborting.')

        func(self, sorted_messages)

    return decorator


def participation_status(
    func: Callable[..., Tuple[Dict[Chat, DataFrame], Dict[str, Any]]]
):
    """
    """

    # Hack to avoid circular imports.
    from .frames import BaseFrame

    def decorator(self: BaseFrame, chats: List[Chat]) -> None:
        modes: Dict[str, Any] = {
            'By Time': generate_chart, 'Treemap': generate_treemap
        }
        choices = list(modes.keys())

        name = select('Choose your mode:', choices).ask()
        sorter_type = modes[name]

        dataframes, kwargs = func(self, chats)
        sorter_type(dataframes, **kwargs)

    return decorator


def group_users(func: Callable[..., Any]):
    """
    """

    # Hack to avoid circular imports.
    from .frames import BaseFrame 

    def decorator(self: BaseFrame, chats: List[Chat]) -> None:
        data: Dict[Chat, Dict[str, List[Message]]] = {}

        for chat in chats:
            messages = {act.display_name: act.messages for act in chat.actors}
            data[chat] = messages

        return func(self, data)

    return decorator



"""
MIT License
Copyright (c) 2021 Qualichat
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from pathlib import Path
from typing import Any, Dict, List, Union

from .utils import log, config, get_random_name
from .models import Actor, Message, SystemMessage
from .regex import CHAT_FORMAT_RE, USER_MESSAGE_RE


__all__ = ('Chat',)


def _clean_impurities(content: str) -> str:
    # Regex will not be used here since
    # :meth:`str.replace` is faster and simpler.
    content = content.replace('\u200e', '')
    content = content.replace('\u2002', '')
    content = content.replace('\u202c', '')
    content = content.replace('\u202a', '')

    content = content.replace('\xa0', ' ')
    content = content.replace('\u2011', '-')

    return content


class Chat:
    """
    """

    __slots__ = ('path', 'filename', 'messages', 'system_messages', '_actors')

    def __init__(self, path: Union[str, Path], **kwargs: Any) -> None:
        if not isinstance(path, Path):
            path = Path(path)

        if not path.is_file():
            raise FileNotFoundError(f'no such file: {str(path)!r}')

        self.path = path.resolve()
        self.filename = self.path.name

        name = f'[green]{str(path)}[/]'
        log('info', f'Loading chat {name}...')

        log('debug', f'Reading {name} file content...')
        encoding = kwargs.pop('encoding', 'utf-8')
        content = path.read_text(encoding=encoding)

        log('debug', f'File {name} read. Cleaning it...')
        raw_data = _clean_impurities(content)

        log('debug', f'Contents of the file {name} cleaned. Parsing it...')
        self.messages: List[Message] = []
        self.system_messages: List[SystemMessage] = []

        self._actors: Dict[str, Actor] = {}

        for match in CHAT_FORMAT_RE.finditer(raw_data):
            if not match:
                # An unknown message? Anyway, 
                # let's just ignore it and move on.
                continue

            created_at = match.group(1)
            rest = match.group(2)

            is_user_message = USER_MESSAGE_RE.match(rest)

            if is_user_message:
                contact_name = is_user_message.group(1)
                content = is_user_message.group(2)

                if contact_name not in self._actors:
                    path = str(self.path)

                    if path not in config:
                        config[path] = {}

                    group: Dict[str, str] = config[path]

                    if contact_name not in group:
                        group[contact_name] = get_random_name()

                    display_name = group[contact_name]
                    self._actors[contact_name] = Actor(display_name)

                actor = self._actors[contact_name]                
                message = Message(actor, content, created_at)

                self.messages.append(message)
                actor.messages.append(message)
            else:
                # It is a system message, indicating some group 
                # event (actor left/joined, changed the group icon,
                # etc.)
                self.system_messages.append(SystemMessage(rest, created_at))

        config.save()

        messages = len(self.messages) + len(self.system_messages)
        actors = len(self.actors)

        msg = f'Loaded {messages:,} messages and {actors:,} actors from {name}.'
        log('info', msg)

    @property
    def actors(self) -> List[Actor]:
        """List[:class:`.Actor`]: The list of actors present in the
        chat.
        """
        return list(self._actors.values())

    def __repr__(self) -> str:
        filename = f'filename={self.filename!r}'
        actors = f'actors={len(self._actors)}'
        messages = f'messages={len(self.messages)}'
        system_messages = f'system_messages={len(self.system_messages)}'

        return f'<Chat {filename} {actors} {messages} {system_messages}>'




"""
MIT License

Copyright (c) 2021 Qualichat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import logging
import os
import random
from pathlib import Path
from typing import Any, Dict, List, Literal
from functools import cache

from tldextract import extract # type: ignore


__all__ = ('config',)


class Config:
    """
    """

    __slots__ = ('data')

    def __init__(self) -> None:
        self.data = self.load()

    def __contains__(self, item: str) -> bool:
        return item in self.data

    def __getitem__(self, item: str) -> Any:
        return self.data[item]

    def __setitem__(self, key: str, value: Any) -> None:
        self.data[key] = value

    def load(self) -> Dict[str, Any]:
        home = Path.home()

        folder = home / '.qualichat'
        config = folder / 'config.json'

        if not folder.is_dir():
            folder.mkdir(exist_ok=True)

            with config.open('w', encoding='utf-8') as file:
                file.write(r'{}')

            return {}

        with config.open('r', encoding='utf-8') as file:
            return json.load(file)

    def save(self) -> None:
        home = Path.home()

        folder = home / '.qualichat'
        config = folder / 'config.json'

        with config.open('w', encoding='utf-8') as file:
            json.dump(self.data, file, indent=2, sort_keys=True)


def _get_all_names() -> List[str]:
    path = os.path.dirname(__file__)
    books = os.path.join(path, 'books.txt')

    with open(books, encoding='utf-8') as f:
        return f.read().split('\n')


def get_random_name() -> str:
    name = random.choice(names)
    # Remove the book from the list so there is no risk that two 
    # actors have the same display name.
    names.remove(name)
    return name.strip()


names = _get_all_names()
config = Config()


logger = logging.getLogger('qualichat')


def log(
    level: Literal['debug', 'info', 'warn', 'error'], message: str
) -> None:
    """Logs a message to ``qualichat`` logger.

    Parameters
    ----------
    level: :class:`str`
        The logging level to send. It must be one of these: `debug`,
        `info`, `warn`, `error`.
    message: :class:`str`
        The message content to send.
    """
    getattr(logger, level)(message)


domains: Dict[str, str] = {
    'youtu.be': 'YouTube',
    'youtube.com': 'YouTube',
    'whatsapp.com': 'WhatsApp',
    't.me': 'Telegram',
    'uol.com.br': 'UOL',
    'glo.bo': 'Globo',
    'bit.ly': 'Bitly',
    'metropoles': 'Metrópoles',
    'theintercept.com': 'The Intercept',
    'estadao.com.br': 'Estadão',
    'diarioonline.com.br': 'Diário Online',
    'brasildefato.com.br': 'Brasil de Fato',
    'ig.com.br': 'IG',
    'terrabrasilnoticias.com': 'Terra Brasil Notícias',
    'folhadapolitica.com': 'Folha da Política',
    'gazetadopovo.com.br': 'Gazeta do Povo',
    'fb.me': 'Facebook',
    'fb.watch': 'Facebook',
    't.co': 'Twitter',
    'goo.gl': 'Google',
    'www.gov.br': 'Governo do Brasil',
}


@cache
def parse_domain(url: str) -> str:
    """Parses a URL to return its sanitized domain.
    Parameters
    ----------
    url: :class:`str`
        The URL to be parsed.
    Returns
    -------
    :class:`str`
        The parsed domain name.
    """
    result = extract(url)

    domain: str = result.domain # type: ignore
    suffix: str = result.suffix # type: ignore

    website = f'{domain}.{suffix}'

    if website in domains:
        return domains[website]

    return domain.capitalize()



"""
MIT License
Copyright (c) 2021 Qualichat
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from enum import Enum, IntEnum


__all__ = ('Period', 'SubPeriod', 'MessageType')


class Period(Enum):
    dawn    = 'Dawn'
    morning = 'Morning'
    evening = 'Evening'
    night   = 'Night'


class SubPeriod(Enum):
    resting            = 'Resting'
    transport_morning  = 'Transport (morning)'
    work_morning       = 'Work (morning)'
    lunch              = 'Lunch'
    work_evening       = 'Work (evening)'
    transport_evening  = 'Transport (evening)'
    second_office_hour = 'Second Office Hour'


class MessageType(IntEnum):
    default              = 0
    gif_omitted          = 1
    image_omitted        = 2
    video_omitted        = 3
    audio_omitted        = 4
    sticker_omitted      = 5
    document_omitted     = 6
    contact_card_omitted = 7
    deleted_message      = 8


def get_period(created_at: datetime.datetime) -> Period:
    if 0 <= created_at.hour < 6:
        period = Period.dawn
    elif 6 <= created_at.hour < 12:
        period = Period.morning
    elif 12 <= created_at.hour < 18:
        period = Period.evening
    else:
        period = Period.night

    return period


def get_sub_period(created_at: datetime.datetime) -> SubPeriod:
    if 0 <= created_at.hour < 6:
        sub_period = SubPeriod.resting
    elif 6 <= created_at.hour < 9:
        sub_period = SubPeriod.transport_morning
    elif 9 <= created_at.hour < 12:
        sub_period = SubPeriod.work_morning
    elif 12 <= created_at.hour < 15:
        sub_period = SubPeriod.lunch
    elif 15 <= created_at.hour < 18:
        sub_period = SubPeriod.work_evening
    elif 18 <= created_at.hour < 21:
        sub_period = SubPeriod.transport_evening
    else:
        sub_period = SubPeriod.second_office_hour

    return sub_period


def get_message_type(content: str) -> MessageType:
    if content == 'image omitted':
        message_type = MessageType.image_omitted
    elif content == 'GIF omitted':
        message_type = MessageType.gif_omitted
    elif content == 'video omitted':
        message_type = MessageType.video_omitted
    elif content == 'audio omitted':
        message_type = MessageType.audio_omitted
    elif content == 'sticker omitted':
        message_type = MessageType.sticker_omitted
    elif content.endswith('document omitted'):
        message_type = MessageType.document_omitted
    elif content == 'Contact card omitted':
        message_type = MessageType.contact_card_omitted
    elif content == 'This message was deleted.':
        message_type = MessageType.deleted_message
    else:
        message_type = MessageType.default

    return message_type



"""
Qualichat
~~~~~~~~~

Open-source linguistic ethnography tool for
framing public opinion in mediatized groups.

:copyright: (c) 2021-present Ernest Manheim
:license: MIT, see LICENSE for more details.
"""

__title__ = 'qualichat'
__author__ = 'Ernest Manheim'
__license__ = 'MIT'
__copyright__ = 'Copyright 2021-present Ernest Manheim'
__version__ = '1.3.8'


__all__ = ('load_chats',)


from typing import NamedTuple

from .core import load_chats as load_chats


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int

version_info = VersionInfo(1, 3, 8, 'stable', 0)



"""
MIT License
Copyright (c) 2021 Qualichat
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from math import sqrt
from collections import defaultdict
import inspect
from pathlib import Path
from types import FunctionType
from typing import (
    Any,
    ClassVar,
    Counter,
    DefaultDict,
    Dict,
    List,
    Optional,
    OrderedDict,
    Pattern,
    Set,
    Tuple,
    Union
)

import spacy
from wordcloud import WordCloud # type: ignore
from pandas import DataFrame
from qualitube import Client # type: ignore
from tldextract import extract # type: ignore

from . import sorters
from .chat import Chat
from ._partials import *
from .enums import MessageType
from .models import Message
from .sorters import generate_treemap, generate_wordcloud, generate_chart, generate_table
from .utils import config, log, parse_domain
from .regex import SHORT_YOUTUBE_LINK_RE, YOUTUBE_LINK_RE


__all__ = ('BaseFrame', 'KeysFrame', 'ParticipationStatusFrame')


nlp = spacy.load('pt_core_news_md')

weekdays = (
    'Monday', 'Tuesday', 'Wednesday', 'Thursday',
    'Friday', 'Saturday', 'Sunday'
)


def _normalize_frame_name(name: str) -> str:
    return name.replace('_', ' ').title()


def _normalize_row(row: List[int], actor: str, chat: Chat) -> List[int]:
    if actor != 'Others':
        return row

    return [int(i / len(chat.actors)) for i in row]


def _parse_nlp(word: str, *, pos: str):
    doc = nlp(word)
    endings = ('ar', 'er', 'ir')

    for token in doc:
        if token.pos_ == pos:
            if pos == 'VERB' and not token.text.endswith(endings):
                continue

            yield token.text


def _parse_nlp_messages(messages: List[Message]):
    types = {'Verbs': 'VERB', 'Nouns': 'NOUN', 'Adjectives': 'ADJ'}

    choices = list(types.keys())
    msg = 'Choose a morphological class:'

    morphological_class = select(msg, choices).ask()
    pos = types[morphological_class]

    with progress_bar() as progress:
        for message in progress.track(messages, description='Parsing...'):
            for text in _parse_nlp(message['Qty_char_text'], pos=pos):
                yield text


def _generate_wordcloud(data: List[str]):
    parent = Path(__file__).resolve().parent
    path = parent / 'fonts' / 'Roboto-Regular.ttf'

    configs: Dict[str, Any] = {}
    configs['width'] = 1920
    configs['height'] = 1080
    configs['font_path'] = str(path)
    configs['background_color'] = 'white'

    all_words = ' '.join(data)
    return WordCloud(**configs).generate(all_words) # type: ignore


class BaseFrame:
    """Represents the base of a Qualichat frame.
    Generally, you should use the built-in frames that Qualichat
    offers you. However, you can subclass this class and create your
    own frames.
    .. note::
        To automatically generate charts, you must create a method
        decorated with :meth:`generate_chart` that returns a
        :class:`pandas.DataFrame`.
    Attributes
    ----------
    chats: List[:class:`.Chat`]
        All the chats loaded via :meth:`qualichat.load_chats`.
    """

    __slots__ = ('chats', 'charts')

    fancy_name: ClassVar[str]

    def __init__(self, chats: List[Chat]) -> None:
        def predicate(o: object) -> bool:
            if inspect.ismethod(o):
                return not o.__name__.startswith('_') # type: ignore

            return False

        methods = inspect.getmembers(self, predicate=predicate)
        charts = {}

        for name, method in methods:
            charts[_normalize_frame_name(name)] = method

        self.chats = chats
        self.charts: Dict[str, FunctionType] = charts

    def __repr__(self) -> str:
        return '<%s>' % self.__class__.__name__


class KeysFrame(BaseFrame):
    """
    """

    fancy_name = 'Keys'

    @sorters.keys
    def keyword(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> None:
        """
        """
        wordclouds: Dict[Chat, WordCloud] = {}
        title = 'Keys Frame (Keyword)'

        result: str = input('Enter the keyword:').ask()
        keyword = result

        for chat, data in chats_data.items():
            new_messages: List[Message] = []

            for messages in data.values():
                for message in messages:
                    if message['Type'] is not MessageType.default:
                        continue

                    if keyword.lower() not in message.content.lower():
                        continue

                    new_messages.append(message)

            messages_data: List[str] = list(_parse_nlp_messages(new_messages))
            wordclouds[chat] = _generate_wordcloud(messages_data)

        generate_wordcloud(wordclouds, title=title)

    @sorters.keys
    def messages(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> None:
        """
        """
        wordclouds: Dict[Chat, WordCloud] = {}
        title = 'Keys Frame (Messages)'

        for chat, data in chats_data.items():
            new_messages: List[Message] = []

            for messages in data.values():
                for message in messages:
                    if message['Type'] is not MessageType.default:
                        continue

                    new_messages.append(message)

            messages_data: List[str] = list(_parse_nlp_messages(new_messages))
            wordclouds[chat] = _generate_wordcloud(messages_data)

        generate_wordcloud(wordclouds, title=title)

    @sorters.keys
    def laminations(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> None:
        """
        """
        title = 'Keys Frame (Laminations)'
        dataframes: Dict[Chat, DataFrame] = {}

        bars = [
            'Qty_char_links', 'Qty_char_emails',
            'Qty_char_calls', 'Qty_char_emoji'
        ]
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[int]] = []

            for actor, messages in data.items():
                links = 0
                emojis = 0
                emails = 0
                calls = 0

                for message in messages:
                    links += len(message['Qty_char_links'])
                    emojis += len(message['Qty_char_emoji'])
                    emails += len(message['Qty_char_emails'])
                    calls += len(message['Qty_char_calls'])

                row = [links, emojis, emails, calls, len(messages)]
                rows.append(_normalize_row(row, actor, chat))

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        generate_chart(dataframes, lines=lines, bars=bars, title=title)

    @sorters.keys
    def links(self, chats_data: Dict[Chat, Dict[str, List[Message]]]) -> None:
        """
        """
        title = 'Keys Frame (Links)'
        dataframes: Dict[Chat, DataFrame] = {}

        bars = ['Qty_char_links']
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[int]] = []

            for actor, messages in data.items():
                links = 0

                for message in messages:
                    links += len(message['Qty_char_links'])

                row = [links, len(messages)]
                rows.append(_normalize_row(row, actor, chat))

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        generate_chart(dataframes, lines=lines, bars=bars, title=title)

    @sorters.keys
    def calls(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> None:
        """
        """
        title = 'Keys Frame (Calls)'
        dataframes: Dict[Chat, DataFrame] = {}

        bars = ['Qty_char_calls']
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[int]] = []

            for actor, messages in data.items():
                calls = 0

                for message in messages:
                    calls += len(message['Qty_char_calls'])

                row = [calls, len(messages)]
                rows.append(_normalize_row(row, actor, chat))

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        generate_chart(dataframes, lines=lines, bars=bars, title=title)

    @sorters.keys
    def emails(self, chats_data: Dict[Chat, Dict[str, List[Message]]]) -> None:
        """
        """
        title = 'Keys Frame (E-mails)'
        dataframes: Dict[Chat, DataFrame] = {}

        bars = ['Qty_char_emails']
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[int]] = []

            for actor, messages in data.items():
                emails = 0

                for message in messages:
                    emails += len(message['Qty_char_emails'])

                row = [emails, len(messages)]
                rows.append(_normalize_row(row, actor, chat))

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        generate_chart(dataframes, lines=lines, bars=bars, title=title)

    @sorters.keys
    def textual_symbols(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> None:
        """
        """
        title = 'Keys Frame (Textual symbols)'
        dataframes: Dict[Chat, DataFrame] = {}

        bars = ['Qty_char_marks', 'Qty_char_emoji']
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[int]] = []

            for actor, messages in data.items():
                marks = 0
                emojis = 0

                for message in messages:
                    marks += len(message['Qty_char_marks'])
                    emojis += len(message['Qty_char_emoji'])

                row = [marks, emojis, len(messages)]
                rows.append(_normalize_row(row, actor, chat))

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        generate_chart(dataframes, lines=lines, bars=bars, title=title)

    @sorters.keys
    def ratings(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> None:
        """
        """
        if not (api_key := config['google_api_key']):
            return log('error', 'No API Key provided. Please provide one.')

        client = Client(api_key)

        title = 'Keys Frame (Ratings)'
        tables: Dict[Chat, DataFrame] = {}

        columns = [
            'Media', 'Actor', 'Date', 'Link',
            'Views', 'Likes', 'Comments', 'Title'
        ]

        regexes: Dict[Tuple[str, str], Pattern[str]] = {
            ('youtu', 'be'): SHORT_YOUTUBE_LINK_RE,
            ('youtube', 'com'): YOUTUBE_LINK_RE
        }

        for chat, data in chats_data.items():
            rows: List[List[Any]] = []
            url_counter: Counter[str] = Counter()

            views_cache: DefaultDict[str, Optional[int]] = defaultdict(int)
            likes_cache: DefaultDict[str, Optional[int]] = defaultdict(int)
            comments_cache: DefaultDict[str, Optional[int]] = defaultdict(int)
            titles_cache: DefaultDict[str, Optional[str]] = defaultdict(str)

            all_messages: List[Message] = []
            
            for messages in data.values():
                all_messages.extend(messages)

            with progress_bar() as progress:
                for message in progress.track(all_messages):
                    created_at = str(message.created_at)
                    actor = message.actor.display_name

                    for url in message['Qty_char_links']:
                        domain = parse_domain(url)

                        if domain == 'YouTube':
                            url_counter[url] += 1

                            # Only repeated items are parsed.
                            if url_counter[url] != 2:
                                continue

                            _, domain, suffix = extract(url) # type: ignore
                            regex = regexes[(domain, suffix)] # type: ignore

                            if not (match := regex.match(url)):
                                continue

                            id = match.group(1)
                            videos = client.get_videos([id]).videos

                            if not videos:
                                continue

                            video = videos[0]

                            views_cache[url] = video.view_count
                            likes_cache[url] = video.like_count
                            comments_cache[url] = video.comment_count
                            titles_cache[url] = video.title

                        views = views_cache[url] or ''
                        likes = likes_cache[url] or ''
                        comments = comments_cache[url] or ''
                        video_title = titles_cache[url]

                        rows.append([
                            domain, actor, created_at, url,
                            views, likes, comments, video_title
                        ])

            dataframe = DataFrame(rows, columns=columns)
            tables[chat] = dataframe

        generate_table(tables, columns=columns, title=title)


class ParticipationStatusFrame(BaseFrame):
    """
    """

    fancy_name = 'Participation Status'

    @sorters.group_users
    def bots(self, chats_data: Dict[Chat, Dict[str, List[Message]]]) -> None:
        """
        """
        title = ''
        choices = ['Index', 'Components']

        msg = 'Choose you action'
        result = select(msg, choices).ask()

        if result == 'Index':
            return _bots_index(chats_data, title)
        else:
            return _bots_components(chats_data, title)

    def messages_per_actors_per_weekday(self, chats: List[Chat]) -> None:
        """
        """
        dataframes: Dict[Chat, DataFrame] = {}
        title = 'Participation Status (Messages per Actors per Weekday)'

        bars = ['Qtd_messages']

        types = {
            'User Messages': 'messages',
            'System Messages': 'system_messages'
        }

        msg = 'Choose the message type:'
        result = select(msg, list(types.keys())).ask()

        message_type = types[result]

        for chat in chats:
            data = OrderedDict({weekday: 0 for weekday in weekdays})

            for message in getattr(chat, message_type):
                data[message.created_at.strftime('%A')] += 1

            index = list(data.keys())
            rows = list(data.values())

            dataframe = DataFrame(rows, index=index, columns=bars)
            dataframes[chat] = dataframe

        generate_chart(dataframes, bars=bars, lines=[], title=title)

    def media_repertoire(self, chats: List[Chat]) -> None:
        """
        """
        title = 'Participation Status (Media Repertoire)'
        choices = ['Choose Media', 'Average Media', 'Treemap']

        msg = 'Choose you action'
        result = select(msg, choices).ask()

        if result == 'Choose Media':
            return _choose_media(chats, title)
        elif result == 'Average Media':
            return _average_media(chats, title)
        else:
            return _media_treemap(chats, title)

    @sorters.group_users
    @sorters.participation_status
    def messages_per_actors(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> Tuple[Dict[Chat, DataFrame], Dict[str, Any]]:
        """
        """
        dataframes: Dict[Chat, DataFrame] = {}
        title = 'Participation Status Frame (Messages per Actors)'

        bars = ['Qty_char_text', 'Qty_char_net', 'Qty_char_total']
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[Union[int, float]]] = []

            for messages in data.values():
                chars_text = 0
                chars_net = 0
                chars_total = 0

                for message in messages:
                    if message['Type'] is not MessageType.default:
                        continue

                    chars_text += len(message['Qty_char_text'].split())
                    chars_net += len(message['Qty_char_net'].split())
                    chars_total += len(message.content.split())

                rows.append([
                    chars_text, chars_net, chars_total, len(messages)
                ])

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        return (dataframes, {'bars': bars, 'lines': lines, 'title': title})

    @sorters.group_users
    @sorters.participation_status
    def message_statistics(
        self, chats_data: Dict[Chat, Dict[str, List[Message]]]
    ) -> Tuple[Dict[Chat, DataFrame], Dict[str, Any]]:
        """
        """
        dataframes: Dict[Chat, DataFrame] = {}
        title = 'Participation Status Frame (Message Statistics)'

        bars = ['Avg_chars_net', 'Avg_chars_text', 'Sd_chars_net']
        lines = ['Qty_messages']

        for chat, data in chats_data.items():
            rows: List[List[Union[int, float]]] = []

            for messages in data.values():
                chars_total = 0
                chars_net: List[int] = []
                chars_text: List[int] = []

                for message in messages:
                    chars_total += len(message.content)
                    chars_net.append(len(message['Qty_char_net']))
                    chars_text.append(len(message['Qty_char_text']))

                average_net = sum(chars_net) / len(chars_net)
                average_text = sum(chars_text) / len(chars_text)

                # SD means "Standard Deviation".
                # See more: https://en.wikipedia.org/wiki/Standard_deviation
                sigma = sum([(x - average_net) ** 2 for x in chars_net])
                sd_net = sqrt(sigma / len(chars_net))

                rows.append([average_net, average_text, sd_net, len(messages)])

            index = list(data.keys())

            dataframe = DataFrame(rows, index=index, columns=bars + lines)
            dataframes[chat] = dataframe

        return (dataframes, {'bars': bars, 'lines': lines, 'title': title})

    @sorters.participation_status
    def laminations_per_actors(
        self, chats: List[Chat]
    ) -> Tuple[Dict[Chat, DataFrame], Dict[str, Any]]:
        """
        """
        title = 'Participation Status (Laminations per Actors)'
        choices = ['Laminations per Actors', 'Average Laminations']

        msg = 'Choose you action'
        result = select(msg, choices).ask()

        if result == 'Laminations per Actors':
            return _laminations_per_actors(chats, title)
        else:
            return _average_laminations(chats, title)

    @sorters.participation_status
    def fabrications_per_actors(
        self, chats: List[Chat]
    ) -> Tuple[Dict[Chat, DataFrame], Dict[str, Any]]:
        """
        """
        title = 'Participation Status (Fabrications per Actors)'
        choices = ['Fabrications per Actors', 'Average Fabrications']

        msg = 'Choose you action'
        result = select(msg, choices).ask()

        if result == 'Machinations per Actors':
            return _fabrications_per_actors(chats, title)
        else:
            return _average_fabrications(chats, title)
        

def _fabrications_per_actors(chats: List[Chat], title: str) -> Any:
    dataframes: Dict[Chat, DataFrame] = {}

    bars = ['Qty_char_laughs', 'Qty_char_marks', 'Qty_char_numbers']
    lines = ['Qty_char_pure']

    for chat in chats:
        rows: List[List[Union[int, float]]] = []

        for actor in chat.actors:
            laughs = 0
            marks = 0
            numbers = 0
            chars_text = 0
            
            for message in actor.messages:
                laughs += len(message['Qty_char_laughs'])
                marks += len(message['Qty_char_marks'])
                numbers += len(message['Qty_char_numbers'])
                chars_text += len(message['Qty_char_text'].split())

            rows.append([
                laughs, marks, numbers, chars_text
            ])

        index = [actor.display_name for actor in chat.actors]

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe

    return (dataframes, {'lines': lines, 'bars': bars, 'title': title})


def _average_fabrications(chats: List[Chat], title: str) -> Any:
    dataframes: Dict[Chat, DataFrame] = {}

    bars = ['Avg_chars_text', 'Sd_chars_text']
    lines = ['Qty_messages']

    for chat in chats:
        rows: List[List[Union[int, float]]] = []

        for actor in chat.actors:
            chars_text: List[int] = []

            for message in actor.messages:
                chars_text.append(len(message['Qty_char_text']))

            average_text = sum(chars_text) / len(chars_text)

            # SD means "Standard Deviation".
            # See more: https://en.wikipedia.org/wiki/Standard_deviation
            sigma = sum([(x - average_text) ** 2 for x in chars_text])
            sd_text = sqrt(sigma / len(chars_text))

            rows.append([average_text, sd_text, len(actor.messages)])

        index = [actor.display_name for actor in chat.actors]

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe

    return (dataframes, {'lines': lines, 'bars': bars, 'title': title})


def _laminations_per_actors(chats: List[Chat], title: str) -> Any:
    dataframes: Dict[Chat, DataFrame] = {}

    bars = [
        'Qty_char_calls', 'Qty_char_links',
        'Qty_char_emails', 'Qty_char_emoji'
    ]
    lines = ['Qty_char_net']

    for chat in chats:
        rows: List[List[Union[int, float]]] = []

        for actor in chat.actors:
            calls = 0
            links = 0
            emails = 0
            emoji = 0
            chars_net = 0
            
            for message in actor.messages:
                calls += len(message['Qty_char_calls'])
                links += len(message['Qty_char_links'])
                emails += len(message['Qty_char_emails'])
                emoji += len(message['Qty_char_emoji'])
                chars_net += len(message['Qty_char_net'].split())

            rows.append([
                calls, links, emails, emoji, chars_net
            ])

        index = [actor.display_name for actor in chat.actors]

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe

    return (dataframes, {'lines': lines, 'bars': bars, 'title': title})


def _average_laminations(chats: List[Chat], title: str) -> Any:
    dataframes: Dict[Chat, DataFrame] = {}

    bars = ['Avg_chars_net', 'Sd_chars_net']
    lines = ['Qty_messages']

    for chat in chats:
        rows: List[List[Union[int, float]]] = []

        for actor in chat.actors:
            chars_net: List[int] = []

            for message in actor.messages:
                chars_net.append(len(message['Qty_char_net']))

            average_net = sum(chars_net) / len(chars_net)

            # SD means "Standard Deviation".
            # See more: https://en.wikipedia.org/wiki/Standard_deviation
            sigma = sum([(x - average_net) ** 2 for x in chars_net])
            sd_net = sqrt(sigma / len(chars_net))

            rows.append([average_net, sd_net, len(actor.messages)])

        index = [actor.display_name for actor in chat.actors]

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe

    return (dataframes, {'lines': lines, 'bars': bars, 'title': title})


def _choose_media(chats: List[Chat], title: str) -> None:
    dataframes: Dict[Chat, DataFrame] = {}
    all_media: Set[str] = set()

    for chat in chats:
        for message in chat.messages:
            for url in message['Qty_char_links']:
                all_media.add(parse_domain(url))

    msg = 'Choose a Media'
    whitelist = checkbox(msg, list(all_media)).ask()

    for chat in chats:
        rows: List[List[int]] = []

        for actor in chat.actors:
            data = {domain: 0 for domain in whitelist}

            for message in actor.messages:
                for url in message['Qty_char_links']:
                    domain = parse_domain(url)

                    if domain not in whitelist:
                        continue

                    data[domain] += 1

            rows.append(list(data.values()))

        index = [actor.display_name for actor in chat.actors]

        dataframe = DataFrame(rows, index=index, columns=whitelist)
        dataframes[chat] = dataframe

    generate_chart(dataframes, bars=whitelist, title=title)    


def _average_media(chats: List[Chat], title: str) -> None:
    dataframes: Dict[Chat, DataFrame] = {}

    bars = ['Qty_average', 'Qty_total']
    lines = ['Qty_messages']

    for chat in chats:
        rows: List[List[Union[int, float]]] = []

        for actor in chat.actors:
            total_urls = 0
            
            for message in actor.messages:
                total_urls += len(message['Qty_char_links'])

            average = total_urls / len(actor.messages)
            rows.append([average, total_urls, len(actor.messages)])

        index = [actor.display_name for actor in chat.actors]

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe

    generate_chart(dataframes, lines=lines, bars=bars, title=title)


def _media_treemap(chats: List[Chat], title: str) -> None:
    dataframes: Dict[Chat, DataFrame] = {}

    urls: DefaultDict[str, int] = defaultdict(int)

    for chat in chats:
        for message in chat.messages:
            for url in message['Qty_char_links']:
                urls[parse_domain(url)] += 1

        index = list(urls.keys())
        rows = list(urls.values())
        
        dataframe = DataFrame(rows, index=index)
        dataframes[chat] = dataframe

    generate_treemap(dataframes, title=title, have_parents=False)


def _bots_index(
    chats_data: Dict[Chat, Dict[str, List[Message]]],
    title: str
) -> None:
    dataframes: Dict[Chat, DataFrame] = {}
    title = 'Participation Status (Bots - Beta)'

    bars = ['Qty_score']
    lines = ['Qty_messages']

    for chat, data in chats_data.items():
        rows: List[List[Union[int, float]]] = []

        for messages in data.values():
            chars_net = 0
            videos = 0
            stickers = 0

            for message in messages:
                if message['Type'] is MessageType.default:
                    chars_net += len(message['Qty_char_net'].split())
                elif message['Type'] is MessageType.video_omitted:
                    videos += 1
                elif message['Type'] is MessageType.sticker_omitted:
                    stickers += 1

            result = ((chars_net * 1) + (videos * 2) + (stickers * 3)) / 6
            rows.append([result, len(messages)])

        index = list(data.keys())

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe
        
    generate_chart(dataframes, bars=bars, lines=lines, title=title)


def _bots_components(
    chats_data: Dict[Chat, Dict[str, List[Message]]],
    title: str
) -> None:
    dataframes: Dict[Chat, DataFrame] = {}
    title = 'Participation Status (Bots - Beta)'

    bars = ['Qty_char_net', 'Qty_videos', 'Qty_stickers']
    lines = ['Qty_messages']

    for chat, data in chats_data.items():
        rows: List[List[Union[int, float]]] = []

        for messages in data.values():
            chars_net = 0
            videos = 0
            stickers = 0

            for message in messages:
                if message['Type'] is MessageType.default:
                    chars_net += len(message['Qty_char_net'].split())
                elif message['Type'] is MessageType.video_omitted:
                    videos += 1
                elif message['Type'] is MessageType.sticker_omitted:
                    stickers += 1

            rows.append([chars_net, videos, stickers, len(messages)])

        index = list(data.keys())

        dataframe = DataFrame(rows, index=index, columns=bars + lines)
        dataframes[chat] = dataframe
        
    generate_chart(dataframes, bars=bars, lines=lines, title=title)




"""
MIT License
Copyright (c) 2021 Qualichat
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from typing import Any, Dict, Iterable, List
from types import MappingProxyType

import emojis # type: ignore

from .regex import *
from .abc import BaseMessage
from .enums import get_message_type, get_period, get_sub_period


__all__ = ('Actor', 'Message', 'SystemMessage')


TIME_FORMAT = r'%d/%m/%y %H:%M:%S'

def parse_time(string: str) -> datetime.datetime:
    """Converts the message creation time string to a
    :class:`datetime.datetime` object.
    
    Parameters
    ----------
    string: :class:`str`
        The message's creation time string.
    Returns
    -------
    :class:`datetime.datetime`
        A :class:`datetime.datetime` object of when the message was
        sent.
    """
    return datetime.datetime.strptime(string, TIME_FORMAT)


def remove_all_incidences(content: str, *iterables: Iterable[str]) -> str:
    for iterable in iterables:
        for incidence in iterable:
            content = content.replace(incidence, '')

    return content


class Actor:
    """Represents an actor in the chat.
    
    Attributes
    ----------
    display_name: :class:`str`
        A representative name for this actor, this name is not
        necessarily the actor's real name.
    messages: List[:class:`.Message`]
        A list containing all the messages that this actor sent in the
        chat.
    """

    __slots__ = ('display_name', 'messages')

    def __init__(self, display_name: str) -> None:
        self.display_name: str = display_name
        self.messages: List[Message] = []

    def __repr__(self) -> str:
        return f'<Actor display_name={self.display_name!r} ' \
               f'messages={len(self.messages)}>'


class Message(BaseMessage):
    """Represents a message sent in the chat by an actor.
    This class has a small linguistic analysis interface for message
    content. To get information about the data from this interface, you
    should use :meth:`object.__getitem__`, e.g. ::
        message['Qty_char_total']
    These are the data currently available:
    - ``Qty_char_total``: :class:`int`
        Shows the total number of characters in the message content.
        This is the same thing as doing: ::
            len(message.content)
    - ``Qty_char_emoji``: List[:class:`str`]
        All unicode emojis present in the message content.
    - ``Qty_char_links``: List[:class:`str`]
        All URLs present in the message content.
    - ``Qty_char_emails``: List[:class:`str`]
        All e-mails present in the message content.
    - ``Qty_char_?``: List[:class:`str`]
        All ``?`` characters present in the message content.
    - ``Qty_char_!``: List[:class:`str`]
        All ``!`` characters present in the message content.
    - ``Qty_char_numbers``: List[:class:`str`]
        All numbers characters present in the message content.
    - ``Qty_char_laughs``: List[:class:`str`]
        All laughs characters present in the message content.
        .. warning::
            This data may not accurate as, computationally, laughs
            are difficult to detect as it does not follow an open
            pattern.
    - ``Qty_char_marks``: List[:class:`str`]
        The union of ``Qty_char_!`` and ``Qty_char_?``.
    - ``Qty_char_call``: List[:class:`str`]
        All calls found in the message content.
        .. versionadded:: 1.3
    - ``Qty_char_net``: :class:`str`
        Represents the net content of the message. This removes all
        calls, URLs, emails and emojis from the message content.
    - ``Qty_char_text``: :class:`str`
        Represents the pure content of the message. This takes the net
        content (via ``Qty_char_net``) and removes laughs, marks and
        numbers from this content.
    - ``Type``: :class:`.MessageType`
        Represents the type of message sent.
        For more information see :class:`.MessageType`.
    - ``Day_period``: :class:`.Period`
        The period of day the message was sent. These are the available
        periods (in 24h format):
        - ``Dawn`` (00:00-05:59)
        - ``Morning`` (06:00-11:59)
        - ``Evening`` (12:00-17:59)
        - ``Night`` (18:00-23:59)
        For more information see :class:`.Period`.
    - ``Day_sub_period``: :class:`.SubPeriod`
        The sub-period of the day the message was sent. These are the
        available periods (in 24-hour format):
        - ``Resting`` (00:00-05:59)
        - ``Transport (morning)`` (06:00-08:59)
        - ``Work (morning)`` (09:00-11:59)
        - ``Lunch`` (12:00-14:59)
        - ``Work (evening)`` (15:00-17:59)
        - ``Transport (evening)`` (18:00-20:59)
        - ``Second Office Hour`` (21:00-23:59)
        For more information see :class:`.SubPeriod`.
    Attributes
    ----------
    actor: :class:`.Actor`
        The actor who sent the message.
    content: :class:`str`
        The content of the message.
    created_at: :class:`datetime.datetime`
        The message's creation time.
    """

    __slots__ = ('actor', 'content', 'created_at', '_data')

    def __init__(self, actor: Actor, content: str, created_at: str) -> None:
        self.actor: Actor = actor
        self.content: str = content
        self.created_at: datetime.datetime = parse_time(created_at)

        data: Dict[str, Any] = {}
        data['Qty_char_total'] = len(self.content)

        all_emojis = list(emojis.iter(self.content)) # type: ignore
        data['Qty_char_emoji'] = all_emojis
        data['Qty_char_links'] = URL_RE.findall(self.content)
        data['Qty_char_emails'] = EMAIL_RE.findall(self.content)

        # We create a copy of the content (since we don't want to
        # change the original content) and then remove all URLs present
        # in the message, to avoid ambiguity.
        content = remove_all_incidences(self.content, *data['Qty_char_links'])

        data['Qty_char_?'] = QUESTION_MARK_RE.findall(content)
        data['Qty_char_!'] = EXCLAMATION_MARK_RE.findall(content)
        data['Qty_char_calls'] = MENTION_RE.findall(content)
        data['Qty_char_numbers'] = NUMBERS_RE.findall(content)
        data['Qty_char_laughs'] = LAUGHS_RE.findall(content)
        data['Qty_char_emoticons'] = EMOTICONS_RE.findall(content)

        all_marks: List[str] = data['Qty_char_emoji'] + data['Qty_char_emoticons']
        data['Qty_char_marks'] = all_marks

        net_incidences_fields = [
            'Qty_char_calls',
            'Qty_char_links',
            'Qty_char_emails',
            'Qty_char_emoji'
        ]

        net_incidendes = [data[i] for i in net_incidences_fields]
        net_text = remove_all_incidences(self.content, *net_incidendes)

        pure_incidences_fields = [
            'Qty_char_laughs',
            'Qty_char_marks',
            'Qty_char_numbers'
        ]

        pure_incidences = [data[i] for i in pure_incidences_fields]
        pure_text = remove_all_incidences(net_text, *pure_incidences)

        data['Qty_char_net'] = net_text
        data['Qty_char_text'] = pure_text

        data['Type'] = get_message_type(self.content)
        data['Day_period'] = get_period(self.created_at)
        data['Day_sub_period'] = get_sub_period(self.created_at)

        self._data: MappingProxyType[str, Any] = MappingProxyType(data)

    def __repr__(self) -> str:
        actor = f'actor={self.actor}'
        created_at = f'created_at={self.created_at!r}'

        return f'<Message {actor} {created_at}>'

    def __getitem__(self, key: Any) -> Any:
        if not isinstance(key, str):
            raise TypeError('indices must be strings')

        return self._data[key]


class SystemMessage(BaseMessage):
    """Represents a message sent in the chat by the system.
    Attributes
    ----------
    content: :class:`str`
        The content of the message.
    created_at: :class:`datetime.datetime`
        The message's creation time.
    """

    __slots__ = ('content', 'created_at')

    def __init__(self, content: str, created_at: str) -> None:
        self.content: str = content
        self.created_at: datetime.datetime = parse_time(created_at)

    def __repr__(self) -> str:
        return f'<SystemMessage created_at={self.created_at!r}>'



import re
from setuptools import setup # type: ignore


# Get library version
VERSION_REGEX = re.compile(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', re.M)

version = ''
with open('qualichat/__init__.py') as f:
    version = VERSION_REGEX.search(f.read()).group(1) # type: ignore

if not version:
    raise RuntimeError('version is not set')


# Get README file
readme = ''
with open('README.md') as f:
    readme = f.read()


# Get requeirements
requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()


setup(
    name='qualichat',
    author='Erneist Manhein',
    project_urls={
        'Documentation': 'https://qualichat.readthedocs.io/en/latest/',
        'Issue tracker': 'https://github.com/qualichat/qualichat/issues',
    },
    version=version,
    packages=['qualichat'],
    license='MIT',
    description="Groups' frames of relevance",
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    python_requires='>=3.7.1',
    package_dir={'qualichat': 'qualichat'},
    package_data={
        'qualichat': ['books.txt'],
        '': ['fonts/*.ttf']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ]
)

